import openai
from rich import print
from tenacity import retry, wait_random_exponential, stop_after_attempt



@retry(wait=wait_random_exponential(multiplier=1, max=40), stop=stop_after_attempt(3))
def chat_completion_request(messages, functions=None, function_call=None, model='gpt-3.5-turbo'):
    call_data = {"model": model, "messages": messages}

    if functions is not None:
        call_data['functions'] = functions

    if function_call is not None:
        call_data['function_call'] = function_call
    try:
        return openai.ChatCompletion.create(**call_data)
    except Exception as e:
        print("Unable to generate ChatCompletion response")
        print(f"Exception: {e}")
        return e



import json
import textwrap
from typing import Callable

from rich import print


from qabot.functions.describe_duckdb_table import describe_table_or_view
from qabot.functions.duckdb_query import run_sql_catch_error
from qabot.functions.wikidata import WikiDataQueryTool
from qabot.llm import chat_completion_request
from qabot.prompts.system import system_prompt


class Agent:
    """
    An Agent is a wrapper around a Language Model that can be used to interact with the LLM.
    Our agent also wraps the functions that can be called by the LLM.
    """

    def __init__(self,
                 database_engine=None,
                 model_name: str = 'gpt-3.5-turbo',
                 allow_wikidata: bool = False,
                 clarification_callback: Callable[[str], str] | None = None,
                 verbose=False):
        """
        Create a new Agent.
        """

        self.model_name = model_name
        self.db = database_engine
        self.verbose = verbose
        self.functions = {
            "clarify": clarification_callback,
            "wikidata": lambda query: WikiDataQueryTool()._run(query),
            "execute_sql": lambda query: run_sql_catch_error(database_engine, query),
            "show_tables": lambda: run_sql_catch_error(database_engine, "show tables"),
            "describe_table": lambda table: describe_table_or_view(database_engine, table),
        }
        self.function_specifications = [
            {
                "name": "execute_sql",
                "description": """Run SQL queries with a local DuckDB database engine. Use for accessing data or any math computation""",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "query": {
                            "type": "string",
                            "description": "DuckDB dialect SQL query. Check the table exists first.",
                        },
                    },
                    "required": ["query"],
                }
            },
            {
                "name": "show_tables",
                "description": "Show the locally available database tables and views",
                "parameters": {
                    "type": "object",
                    "properties": {},
                },
            },
            {
                "name": "describe_table",
                "description": "Show the column names and types of a local database table or view",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "table": {
                            "type": "string",
                            "description": "The table or view name",
                        },
                    },
                    "required": ["table"],
                },
            },
            # A special function to call to summarize the answer
            {
                "name": "answer",
                "description": "Final reply to the user question with a detailed fact based answer",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "summary": {
                            "type": "string",
                            "description": "A standalone one-line summary answering the users question",
                        },
                        "detail": {
                            "type": "string",
                            "description": "detailed answer to the users question including how it was computed. Markdown is acceptable",
                        },
                        "query": {
                            "type": "string",
                            "description": "If the user can re-run the query, include it here",
                        },
                        # "value": {
                        #     "type": "any",
                        #     "description": "If the answer is a number or array, include the value here",
                        # }
                    },
                    "required": ["summary", "detail"],
                },
            },

        ]

        if allow_wikidata:
            self.function_specifications.append(
                {
                    "name": "wikidata",
                    "description": textwrap.dedent(
                        """Useful for when you need specific data from Wikidata.
                        Input to this tool is a single correct SPARQL statement for Wikidata. Limit all requests to 10 or fewer rows. 

                        Output is the raw response in json. If the query is not correct, an error message will be returned. 
                        If an error is returned, you may rewrite the query and try again. If you are unsure about the response
                        you can try rewrite the query and try again. Prefer local data before using this tool.
                        """),
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "query": {
                                "type": "string",
                                "description": "A valid SPARQL query for Wikidata",
                            },
                        }
                    }
                }
            )

        if clarification_callback is not None:
            self.function_specifications.append(
                {
                    "name": "clarify",
                    "description": textwrap.dedent(
                        """Useful for when you need to ask the user a question to clarify their request.
                        Input to this tool is a single question for the user. Output is the user's response
                        """),
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "clarification": {
                                "type": "string",
                                "description": "A question or prompt for the user",
                            },
                        }
                    }
                }
            )

        messages = []

        messages.append({"role": "system",
                         "content": system_prompt})
        # messages.append({"role": "user", "content": user_prompt})

        # Let's force the assistant to get the current tables
        messages.append(
            {
                'role': 'assistant',
                'content': None,
                'function_call': {'name': 'show_tables', 'arguments': '{}'}
            }
        )
        messages.append(
            {
                "role": "function",
                "name": 'show_tables',
                "content": execute_function_call(messages[-1], self.functions)
            })

        self.messages = messages

    def __call__(self, user_input):
        """
        Pass new input from the user to the LLM and return the response.
        """
        return self.run(user_input)

    def run(self, user_input):
        """
        Run the LLM/function execution loop and return the final response.
        """
        self.messages.append(
            {
                "role": "user",
                "content": user_input
            })

        for _ in range(50):
            chat_response = chat_completion_request(
                self.messages,
                functions=self.function_specifications,
                model=self.model_name
            )

            #print("Raw output")
            #print(chat_response)

            message = chat_response.choices[0]['message']
            self.messages.append(message)
            if 'function_call' in message:
                function_name = message["function_call"]["name"]

                results = execute_function_call(message, self.functions)
                print(f"[pink]{results}[/pink]")
                if function_name == 'answer':
                    return results

                else:
                    # Inject a response message for the function call
                    self.messages.append(
                        {
                            "role": "function",
                            "name": message['function_call']['name'],
                            "content": results
                        })


def create_agent_executor(**kwargs):
    return Agent(**kwargs)


def execute_function_call(message, functions):
    function_name = message["function_call"]["name"]
    kwargs = json.loads(message["function_call"]["arguments"])
    if function_name in functions:
        f = functions[function_name]
        results = f(**kwargs)
    elif function_name == 'answer':
        return kwargs
    else:
        results = f"Error: function {message['function_call']['name']} does not exist"
    return results



from typing import Optional

from qabot.agent import Agent
from qabot.duckdb_manual_data_loader import create_duckdb, import_into_duckdb_from_files


def ask_wikidata(query: str, verbose=False):
    agent = Agent(allow_wikidata=True, verbose=verbose)
    result = agent(query)
    return result['summary']


def ask_file(query: str, filename: Optional[str], verbose=False):
    engine = create_duckdb()
    database_engine, executed_sql = import_into_duckdb_from_files(engine, [filename])
    agent = Agent(
        database_engine=database_engine,
        verbose=verbose
    )
    result = agent(query)
    return result['summary']



from qabot.functions.duckdb_query import run_sql_catch_error


def describe_table_or_view(database, table):
    table_columns_and_types_query = f"select column_name, data_type from information_schema.columns where table_name='{table}';"
    table_first_rows_query = f"select * from '{table}' limit 3;"
    table_count_rows_query = f"select count(*) from '{table}';"
    table_description = run_sql_catch_error(database, table_columns_and_types_query)
    table_size = run_sql_catch_error(database, table_count_rows_query)
    table_preview = run_sql_catch_error(database, table_first_rows_query)[:4000]
    return f"{table}\n{table_description}\n{table_count_rows_query}\n{table_size}\n{table_first_rows_query}\n{table_preview}"



from typing import List, Optional
from pydantic import AnyUrl
from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    OPENAI_API_KEY: str

    QABOT_DATABASE_URI: Optional[str] = None
    QABOT_CACHE_DATABASE_URI: AnyUrl = "duckdb:///:memory:"
    QABOT_MODEL_NAME: str = "gpt-4"
    QABOT_TABLES: Optional[List[str]] = None
    QABOT_ENABLE_WIKIDATA: bool = True
    QABOT_ENABLE_HUMAN_CLARIFICATION: bool = True



from rich import print


def pretty_print_conversation(messages):
    role_to_color = {
        "system": "red",
        "user": "green",
        "assistant": "blue",
        "function": "magenta",
    }

    for message in messages:
        color = role_to_color[message["role"]]
        m = f"{message['role']}: {message['content']}"

        if message["role"] == "function":
            m = f"function ({message['name']}): {message['content']}"
        elif message["role"] == "assistant" and "function_call" in message:
            m = f"assistant {message['function_call']['name']}({message['function_call']['arguments']})"
        print(f"[{color}]{m}[/{color}]\n")



system_prompt = """You are Qabot, a large language model trained to interact with DuckDB.
Qabot is designed to be able to assist with a wide range of tasks, from answering simple questions to 
providing in-depth explorations on a wide range of topics relating to data.

Qabot answers questions by first querying for data to guide its answer. Qabot responds with clarifying
questions if the request isn't clear. Qabot prefers to give factual answers backed by data, even
calculations are computed by executing SQL. 

Qabot prefers to split questions into small discrete steps, communicating the plan to the user at each step.

Qabot does NOT make any DML statements (INSERT, UPDATE, DELETE, DROP etc.) to the database.

Unless the user specifies in their question a specific number of examples to obtain, limit any
select query to returning 5 results.

Pay attention to use only the column names that you can see in the schema description. Pay attention
to which column is in which table.
    
If the question does not seem related to the database, Qabot returns "I don't know" as the answer.
"""



import os
from typing import Tuple
from urllib.parse import urlparse
import tempfile

import duckdb
from duckdb import ParserException, ProgrammingError
import requests


def uri_validator(x):
    try:
        result = urlparse(x)
        return all([result.scheme, result.netloc])
    except:
        return False


def create_duckdb(duckdb_path: str = ':memory:') -> duckdb.DuckDBPyConnection:
    # By default, duckdb is fully in-memory - we can provide a path to get
    # persistent storage

    duckdb_connection = duckdb.connect(duckdb_path)
    try:
        duckdb_connection.sql("INSTALL httpfs;")
        duckdb_connection.sql("LOAD httpfs;")
    except Exception:
        print("Failed to install httpfs extension. Loading remote files will not be supported")


    duckdb_connection.sql("create table if not exists qabot_queries(query VARCHAR, timestamp TIMESTAMP DEFAULT CURRENT_TIMESTAMP);")

    return duckdb_connection


def import_into_duckdb_from_files(duckdb_connection: duckdb.DuckDBPyConnection, files: list[str]) -> Tuple[duckdb.DuckDBPyConnection, list[str]]:

    executed_sql = []
    for i, file_path in enumerate(files, 1):

        if file_path.startswith("postgresql://"):
            try:
                duckdb_connection.sql("INSTALL postgres_scanner;")
                duckdb_connection.sql("LOAD postgres_scanner;")
            except Exception:
                print("Failed to install postgres_scanner extension. Loading directly from postgresql will not be supported")
                continue
            duckdb_connection.execute(f"CALL postgres_attach('{file_path}')")
        elif file_path.endswith(".sqlite"):
            try:
                duckdb_connection.sql("INSTALL sqlite;")
                duckdb_connection.sql("LOAD sqlite;")
            except Exception:
                print("Failed to install sqlite extension. Loading directly from sqlite will not be supported")
                continue
            duckdb_connection.execute(f"CALL sqlite_attach('{file_path}')")
        else:
            executed_sql.append(load_external_data_into_db(duckdb_connection, file_path))

    return duckdb_connection, executed_sql


def load_external_data_into_db(conn: duckdb.DuckDBPyConnection, file_path, allow_view=True):
    # Work out if the filepath is actually a url (e.g. s3://)
    is_url = uri_validator(file_path)
    # Get the file name without extension from the file_path
    table_name, extension = os.path.splitext(os.path.basename(file_path))
    # If the table_name isn't a valid SQL identifier, we'll need to use something else

    table_name = table_name.replace("-", "_").replace(".", "_").replace(" ", "_").replace('/', '_')

    try:
        conn.sql(f"create table t_{table_name} as select 1;")
        conn.sql(f"drop table t_{table_name};")
    except (ParserException, ProgrammingError) as e:
        table_name = "data"

    # try to create a view then fallback to a table if it fails
    use_view = allow_view and is_url
    try:
        create_statement = f"create {'view' if use_view else 'table'} '{table_name}' as select * from '{file_path}';"
        conn.sql(create_statement)
    except duckdb.IOException as e:
        # This can occur if the server doesn't send Content-Length headers
        # We can work around this by downloading the data locally and then
        # loading it from there. We assume CSV files for now
        if is_url:

            with tempfile.NamedTemporaryFile(mode='w+b', suffix='.csv', delete=False) as f:
                r = requests.get(file_path, stream=True)
                for chunk in r.iter_content(chunk_size=1024):
                    if chunk:
                        f.write(chunk)
                f.flush()

                create_statement = f"create table '{table_name}' as select * from read_csv_auto('{f.name}');"

        conn.sql(create_statement)

    return create_statement




from typing import List, Optional
import warnings

import typer
from rich import print
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.prompt import Confirm, Prompt

from qabot.config import Settings
from qabot.duckdb_manual_data_loader import import_into_duckdb_from_files, create_duckdb
from qabot.agent import create_agent_executor, Agent

warnings.filterwarnings("ignore")

INITIAL_NON_INTERACTIVE_PROMPT = "🚀 How can I help you explore your database?"
INITIAL_INTERACTIVE_PROMPT = "[bold green] 🚀 How can I help you explore your database?"
FOLLOW_UP_PROMPT = "[bold green] 🚀 anything else I can help you with?"
DUCK_PROMPT = "[bold green] 🦆"

app = typer.Typer(
    pretty_exceptions_show_locals=False,
    pretty_exceptions_enable=True
)



@app.command()
def main(
        query: str = typer.Option("Describe the tables", '-q', '--query', prompt=INITIAL_NON_INTERACTIVE_PROMPT),
        file: Optional[List[str]] = typer.Option(None, "-f", "--file", help="File or url containing data to load and query"),
        database_uri: Optional[str] = typer.Option(":memory:", "-d", "--database", help="DuckDB Database URI (e.g. '/tmp/qabot.duckdb')"),
        enable_wikidata: bool = typer.Option(False, "-w", "--wikidata", help='Allow querying from wikidata'),
        verbose: bool = typer.Option(False, "-v", "--verbose", help='Essentially debug output'),
):
    """
    Query a database or Wikidata using a simple natural language query.

    Example:
        qabot -q "What is the average length of song by artist 'The Beatles'?" -f data/chinook.sqlite
    """

    settings = Settings()
    executed_sql = ''
    # If files are given load data into local DuckDB
    database_engine = create_duckdb(database_uri)

    if len(file) > 0:
        if isinstance(file, str):
            file = [file]
        print("[red]🦆[/red] [bold]Loading data...[/bold]")
        database_engine, executed_sql = import_into_duckdb_from_files(database_engine, file)
        executed_sql = '\n'.join(executed_sql)
    else:
        print("[red]🦆[/red]")

    with Progress(
        SpinnerColumn(),
        TextColumn("[green][progress.description]{task.description}"),
        transient=False,
    ) as progress:

        t2 = progress.add_task(description="Creating LLM agent", total=None)

        def clarification(clarification):
            progress.stop()
            return Prompt.ask(f"[bold red]Clarification requested:[/]\n[bold yellow]{clarification}")

        agent = Agent(
            database_engine=database_engine,
            verbose=False,
            model_name=settings.QABOT_MODEL_NAME,
            allow_wikidata=settings.QABOT_ENABLE_WIKIDATA and enable_wikidata,
            clarification_callback=clarification if settings.QABOT_ENABLE_HUMAN_CLARIFICATION else None,
        )

        progress.remove_task(t2)

        chat_history = [f"""
        Startup SQL Queries:
        ```
        {executed_sql}
        ```
        """]

        while True:

            t = progress.add_task(description="Processing query...", total=None)
            print("[bold red]Query: [/][green]" + query)

            inputs = query

            result = agent(inputs)

            progress.remove_task(t)

            #print("Total tokens", output_callback.total_tokens, f"approximate cost in USD: {openai_callback.total_cost}")

            # Stop the progress before outputting result and prompting for any more input
            progress.stop()
            print()

            print("[bold blue]Question:[/]\n[bold blue]" + query + "\n")
            print("[bold blue]Answer:[/]\n[bold blue]" + result['summary'] + "\n")
            print("\n[blue]" + result['detail'] + "[/blue]\n")
            if 'query' in result:
                print("[cyan]" + result['query'] + "[/cyan]\n")
            chat_history.append(result['summary'])


            print()
            query = Prompt.ask(FOLLOW_UP_PROMPT)

            if query.lower() in {'q', 'exit', 'quit'} and Confirm.ask("Are you sure you want to Quit?"):
                break

            progress.start()


def run():
    app()


if __name__ == '__main__':
    run()
